use crate::prelude::*;

use super::{
    super::{
        super::shared::{bufferhandler::*, safecommandbuffer::SafeCommandBuffer},
        create_shadow_buffer_handler,
        rasterizershader::RasterizerShader,
    },
    ExtraInfo, RasterizerLightInterface,
};

use crate::update_iter;

use anyhow::Result;
use paste;
use utilities::{
    impl_reprc,
    prelude::cgmath::{vec3, Matrix4, Vector3, Vector4},
};

use std::sync::{Arc, Mutex};

impl_reprc!(
    #[derive(Copy)]
    pub struct ExtendedDirectionalLightData {
        data: DirectionalLightData,
        #[assume_reprc]
        view: Matrix4<f32>,
        #[assume_reprc]
        light_pos: Vector4<f32>,
    }
);

pub struct DirectionalLight {
    light_data: Light,
    light_info_buffer: Arc<Buffer<ExtendedDirectionalLightData>>,
    extra_info_buffer: Arc<Buffer<ExtraInfo>>,
    _descriptor_set: Arc<DescriptorSet>,

    view: View,
    buffer_handler: BufferHandler,
    render_target: RenderTarget,
}

impl DirectionalLight {
    #[inline]
    fn create_projection(distance: f32) -> Matrix4<f32> {
        let size = 15.0;

        ortho(-size, size, -size, size, 0.1, distance * 1.5)
    }
}

impl RasterizerLightInterface<ExtendedDirectionalLightData> for DirectionalLight {
    fn new(
        mut light: Light,
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        image_count: usize,
        shadow_image_size: u32,
    ) -> Result<Self> {
        // create render target with an R32 float value
        // for precise depth values
        let render_target = RenderTarget::builder()
            .add_sub_pass(
                SubPass::builder(shadow_image_size, shadow_image_size)
                    .add_target_info(CustomTarget {
                        usage: VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT | VK_IMAGE_USAGE_SAMPLED_BIT,
                        format: VK_FORMAT_R32_SFLOAT,
                        clear_on_load: true,
                        store_on_save: true,
                        attach_sampler: true,
                        use_as_input: false,
                        clear_value: ClearValue::Color([0.0, 0.0, 0.0, 1.0]),
                    })
                    .add_target_info(CustomTarget {
                        usage: VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT
                            | VK_IMAGE_USAGE_TRANSFER_SRC_BIT,
                        format: VK_FORMAT_D16_UNORM,
                        clear_on_load: true,
                        store_on_save: true,
                        attach_sampler: false,
                        use_as_input: false,
                        clear_value: ClearValue::Depth(1.0, 0),
                    })
                    .use_queue(queue.clone())
                    .build(device)?,
            )
            .build(device)?;

        // create pipelines with all bindings to different object types
        let mut buffer_handler = create_shadow_buffer_handler(
            device,
            queue,
            image_count,
            render_target.render_pass().clone(),
            &RasterizerShader::directional_shadow(device)?,
        )?;

        // setup view
        let mut view = View::new(device, Self::create_projection(light.power()), 45.0, true)?;

        let center = vec3(0.0, 0.0, 0.0);

        view.camera_mut().set_center(center);
        view.camera_mut()
            .set_eye_offset(-light.direction() * light.power());
        view.update_buffer()?;

        // create buffer with light info
        let light_info_buffer = {
            let view_matrix = view.camera_mut().view_matrix();

            Buffer::builder()
                .set_usage(VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT)
                .set_memory_usage(MemoryUsage::CpuOnly)
                .set_data(&[ExtendedDirectionalLightData {
                    data: DirectionalLightData::from(&light),
                    view: view.proj() * view_matrix,
                    light_pos: (view.camera().center() + view.camera().eye_offset()).extend(1.0),
                }])
                .build(device.clone())?
        };

        // create buffer with additional information
        let extra_info_buffer = Buffer::builder()
            .set_usage(VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT)
            .set_memory_usage(MemoryUsage::CpuOnly)
            .set_data(&[ExtraInfo { center }])
            .build(device.clone())?;

        // create descriptor set with infos required for shadow mapping
        let layout = DescriptorSetLayout::builder()
            .add_layout_binding(
                0,
                VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER,
                VK_SHADER_STAGE_FRAGMENT_BIT,
                0,
            )
            .add_layout_binding(
                1,
                VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER,
                VK_SHADER_STAGE_FRAGMENT_BIT,
                0,
            )
            .build(device.clone())?;

        let pool = DescriptorPool::builder()
            .set_layout(layout)
            .build(device.clone())?;

        let descriptor_set = DescriptorPool::prepare_set(&pool).allocate()?;

        descriptor_set.update(&[
            DescriptorWrite::uniform_buffers(0, &[&light_info_buffer]),
            DescriptorWrite::uniform_buffers(1, &[&extra_info_buffer]),
        ])?;

        // set update callback, to keep data in sync
        let buffer_clone = light_info_buffer.clone();

        light.set_update_callback(Box::new(move |parameter, view: &mut View| {
            let mut mapped = buffer_clone.map_complete()?;

            match parameter {
                LightParameter::Direction(direction) => {
                    mapped[0].data.direction = direction;
                    view.camera_mut()
                        .set_eye_offset(-direction * mapped[0].data.power);

                    let view_matrix = view.camera_mut().view_matrix();
                    mapped[0].view = view.proj() * view_matrix;

                    view.update_buffer()?;
                }
                LightParameter::Color(color) => mapped[0].data.color = color,
                LightParameter::Power(power) => {
                    mapped[0].data.power = power;
                    view.camera_mut()
                        .set_eye_offset(power * -mapped[0].data.direction.xyz());
                    view.set_proj(Self::create_projection(power));

                    let view_matrix = view.camera_mut().view_matrix();
                    mapped[0].view = view.proj() * view_matrix;

                    view.update_buffer()?;
                }
                _ => (),
            }

            Ok(())
        }));

        // create rendering functions
        update_iter!(
            buffer_handler,
            (position, PositionOnly),
            (position_coloruv, PositionColorUV),
            (position_normal, PositionNormal),
            (position_normal_coloruv, PositionNormalColorUV),
            (position_normal_coloruv_bone, PositionNormalColorUVBone),
            (
                position_normal_coloruv_normaluv,
                PositionNormalColorUVNormalUV
            ),
            (
                position_normal_coloruv_normaluv_bone,
                PositionNormalColorUVNormalUVBone
            ),
        );

        Ok(DirectionalLight {
            light_data: light,
            light_info_buffer,
            extra_info_buffer,
            _descriptor_set: descriptor_set,

            view,
            buffer_handler,
            render_target,
        })
    }

    fn shadow_map(&self) -> &Arc<Image> {
        self.render_target.sub_pass(0).attachments()[0].image(0)
    }

    fn buffer(&self) -> &Arc<Buffer<ExtendedDirectionalLightData>> {
        &self.light_info_buffer
    }

    fn invalidate(&mut self) {
        self.buffer_handler.invalidate();
    }

    fn render(
        &mut self,
        _content: &Vec<&EntityObject>,
        buffer_recorder: &mut CommandBufferRecorder<'_>,
        _index: usize,
    ) -> Result<()> {
        self.render_target.begin(
            buffer_recorder,
            VK_SUBPASS_CONTENTS_SECONDARY_COMMAND_BUFFERS,
            0,
        );

        // let mut list = Vec::new();

        // self.buffer_handler.execute(
        //     Some(self.render_target.framebuffer(0)),
        //     content,
        //     &[&self.descriptor_set],
        //     &mut list,
        //     index,
        //     &self.view,
        // )?;

        println!("directional light renderer is not implemented!");

        // buffer_recorder.execute_commands(&list);

        self.render_target.end(buffer_recorder);

        Ok(())
    }

    fn set_center(&mut self, center: Vector3<f32>) -> Result<()> {
        if self.view.camera().center() != center {
            self.view.camera_mut().set_center(center);

            {
                let mut mapped = self.light_info_buffer.map_complete()?;

                let view_matrix = self.view.camera_mut().view_matrix();
                let view = self.view.proj() * view_matrix;
                let position = self.view.camera().center() + self.view.camera().eye_offset();

                mapped[0].view = view;
                mapped[0].light_pos = position.extend(1.0);
            }

            {
                let mut mapped = self.extra_info_buffer.map_complete()?;
                mapped[0].center = center;
            }

            self.view.update_buffer()?;
        }

        Ok(())
    }

    fn eq(&self, light: &Light) -> bool {
        self.light_data.eq(light)
    }
}

impl Drop for DirectionalLight {
    fn drop(&mut self) {
        self.light_data.clear_update_callback();
    }
}
