use super::super::shared::shadercompiler::{ShaderCompiler, ShaderTypeConverter, Shaders};

use crate::prelude::*;

use std::sync::Arc;

use anyhow::Result;

pub struct RasterizerShader {}

impl RasterizerShader {
    pub fn directional_shadow(device: &Arc<Device>) -> Result<Shaders> {
        let vs_text = include_str!("../shaders/rasterizer/multiple_render_targets.vert");
        let fs_text = include_str!("../shaders/rasterizer/directional_shadow.frag");

        let stages = vec![
            (vs_text, "multiple_render_targets.vert"),
            (fs_text, "directional_shadow.frag"),
        ];

        let compiler = ShaderCompiler::new(stages)?;

        compiler.compile_stages(
            false,
            device,
            &[("CUBE_SHADOW", None)],
            [
                &[],
                &["COLOR_UVS"],
                &["NORMALS"],
                &["NORMALS", "BONES"],
                &["NORMALS", "COLOR"],
                &["NORMALS", "COLOR_UVS"],
                &["NORMALS", "COLOR_UVS", "BONES"],
                &["NORMALS", "COLOR_UVS", "NORMAL_UVS"],
                &["NORMALS", "COLOR_UVS", "NORMAL_UVS", "BONES"],
            ],
        )
    }

    pub fn single_stage(device: &Arc<Device>, shadowed: bool) -> Result<Shaders> {
        let vs_text = include_str!("../shaders/rasterizer/multiple_render_targets.vert");
        let fs_text = include_str!("../shaders/rasterizer/fallback.frag");

        let stages = vec![
            (vs_text, "multiple_render_targets.vert"),
            (fs_text, "fallback.frag"),
        ];

        let compiler = ShaderCompiler::new(stages)?;

        let general = if shadowed {
            vec![("SHADOWED", None)]
        } else {
            Vec::new()
        };

        compiler.compile_stages(
            false,
            device,
            &general,
            [
                &[],
                &["COLOR_UVS"],
                &["NORMALS"],
                &["NORMALS", "BONES"],
                &["NORMALS", "COLOR"],
                &["NORMALS", "COLOR_UVS"],
                &["NORMALS", "COLOR_UVS", "BONES"],
                &["NORMALS", "COLOR_UVS", "NORMAL_UVS"],
                &["NORMALS", "COLOR_UVS", "NORMAL_UVS", "BONES"],
            ],
        )
    }

    pub fn mrt(device: &Arc<Device>) -> Result<Shaders> {
        let vs_text = include_str!("../shaders/rasterizer/multiple_render_targets.vert");
        let fs_text = include_str!("../shaders/rasterizer/multiple_render_targets.frag");

        let stages = vec![
            (vs_text, "multiple_render_targets.vert"),
            (fs_text, "multiple_render_targets.frag"),
        ];

        let compiler = ShaderCompiler::new(stages)?;

        compiler.compile_stages(
            false,
            device,
            &[],
            [
                &[],
                &["COLOR_UVS"],
                &["NORMALS"],
                &["NORMALS", "BONES"],
                &["NORMALS", "COLOR"],
                &["NORMALS", "COLOR_UVS"],
                &["NORMALS", "COLOR_UVS", "BONES"],
                &["NORMALS", "COLOR_UVS", "NORMAL_UVS"],
                &["NORMALS", "COLOR_UVS", "NORMAL_UVS", "BONES"],
            ],
        )
    }

    pub fn traditional(
        device: &Arc<Device>,
        use_shadow_maps: bool,
        point_light_count: u32,
        directional_light_count: u32,
        spot_light_count: u32,
    ) -> Result<Shaders> {
        let vs_text = include_str!("../shaders/rasterizer/traditional.vert");
        let fs_text = include_str!("../shaders/rasterizer/traditional.frag");

        let stages = vec![(vs_text, "traditional.vert"), (fs_text, "traditional.frag")];

        let compiler = ShaderCompiler::new(stages)?;

        let mut default = vec![
            ("POINT_LIGHT_COUNT", Some(point_light_count.to_string())),
            (
                "DIRECTIONAL_LIGHT_COUNT",
                Some(directional_light_count.to_string()),
            ),
            ("SPOT_LIGHT_COUNT", Some(spot_light_count.to_string())),
            ("BONE_COUNT", Some(asset::BONE_COUNT.to_string())),
        ];

        if use_shadow_maps {
            default.push(("SHADOW_MAPS", None));
        }

        compiler.compile_stages(
            false,
            device,
            default.as_slice(),
            [
                &[],
                &["COLOR_UVS"],
                &["NORMALS"],
                &["NORMALS", "BONES"],
                &["NORMALS", "COLOR"],
                &["NORMALS", "COLOR_UVS"],
                &["NORMALS", "COLOR_UVS", "BONES"],
                &["NORMALS", "COLOR_UVS", "NORMAL_UVS"],
                &["NORMALS", "COLOR_UVS", "NORMAL_UVS", "BONES"],
            ],
        )
    }

    pub fn deferred(
        device: &Arc<Device>,
        ambient: f32,
    ) -> Result<(
        Arc<ShaderModule<shader_type::Vertex>>,
        Arc<ShaderModule<shader_type::Fragment>>,
    )> {
        let vs_text = include_str!("../shaders/rasterizer/deferred.vert");
        let fs_text = include_str!("../shaders/rasterizer/deferred.frag");

        let compiler =
            ShaderCompiler::new(vec![(vs_text, "deferred.vert"), (fs_text, "deferred.frag")])?;

        let vertex =
            compiler.compile::<_, 0>(device, &[("AMBIENT", Some(ambient.to_string()))], |_| {})?;
        let fragment =
            compiler.compile::<_, 1>(device, &[("AMBIENT", Some(ambient.to_string()))], |_| {})?;

        Ok((vertex, fragment))
    }

    pub fn deferred_fallback(
        device: &Arc<Device>,
        ambient: f32,
        spot_light_count: u32,
        directional_light_count: u32,
        point_light_count: u32,
    ) -> Result<(
        Arc<ShaderModule<shader_type::Vertex>>,
        Arc<ShaderModule<shader_type::Fragment>>,
    )> {
        let vs_text = include_str!("../shaders/rasterizer/deferred.vert");
        let fs_text = include_str!("../shaders/rasterizer/deferred.frag");

        let compiler =
            ShaderCompiler::new(vec![(vs_text, "deferred.vert"), (fs_text, "deferred.frag")])?;

        let options: &[(&str, Option<String>)] = &[
            ("AMBIENT", Some(ambient.to_string())),
            ("FALLBACK", None),
            ("SPOT_LIGHT_COUNT", Some(spot_light_count.to_string())),
            (
                "DIRECTIONAL_LIGHT_COUNT",
                Some(directional_light_count.to_string()),
            ),
            ("POINT_LIGHT_COUNT", Some(point_light_count.to_string())),
        ];

        let vertex = compiler.compile::<_, 0>(device, options, |_| {})?;
        let fragment = compiler.compile::<_, 1>(device, options, |_| {})?;

        Ok((vertex, fragment))
    }
}
