use crate::prelude::*;

use anyhow::Result;
use utilities::{impl_reprc, prelude::cgmath::Vector4};

use super::{
    super::shared::safecommandbuffer::{ComputeShaderUnit, SafeCommandBuffer},
    animator::VertexData,
    raytraceshader::RayTracingShader,
    RenderingCore,
};

use std::sync::{Arc, Mutex};

impl_reprc!(
    #[derive(Copy)]
    struct ASType {
        #[assume_reprc]
        _bb_min: Vector4<f32>,
        #[assume_reprc]
        _bb_max: Vector4<f32>,

        #[assume_reprc]
        _index_info: Vector4<u32>,
    }
);

impl_reprc!(
    #[derive(Copy)]
    struct SceneInfo {
        #[assume_reprc]
        bvh_node_count: u32,
    }
);

pub struct SoftwareRayTracer {
    raytrace_descriptor: Arc<DescriptorSet>,
    scene_descriptor: Arc<DescriptorSet>,
    as_descriptor: Arc<DescriptorSet>,

    info_buffer: Arc<Buffer<SceneInfo>>,

    // rendering variables
    triangles_per_as: u32,
    width: u32,
    height: u32,
    valid: bool,

    // acceleration structure
    as_buffer: Arc<Buffer<ASType>>,

    // raytracing handles
    raytrace_commands: SafeCommandBuffer<VertexData>,
}

impl RenderingCore for SoftwareRayTracer {
    fn new(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        raytrace_info: RaytracingInfo,
        view: &View,
        image: &Arc<Image>,
        position_image: &Arc<Image>,
        descriptors: Vec<&dyn VkHandle<VkDescriptorSetLayout>>,
    ) -> Result<Self> {
        let (info_buffer, raytrace_descriptor) = Self::create_ray_trace_descriptor(device)?;
        let scene_descriptor = Self::create_scene_descriptor(device)?;
        let as_descriptor = Self::create_scene_descriptor(device)?;

        let as_buffer =
            Self::create_as_buffer(device, raytrace_info.accelerate_buffer_size as VkDeviceSize)?;

        as_descriptor.update(&[DescriptorWrite::storage_buffers(0, &[&as_buffer])])?;

        raytrace_descriptor.update(&[
            DescriptorWrite::storage_buffers(0, &[&as_buffer]),
            DescriptorWrite::storage_images(1, &[&image]),
            DescriptorWrite::storage_images(2, &[&position_image]),
            DescriptorWrite::uniform_buffers(3, &[view.buffer()]),
        ])?;

        let mut descs = Vec::with_capacity(descriptors.len() + 1);
        descs.push(&raytrace_descriptor as &dyn VkHandle<VkDescriptorSetLayout>);

        for descriptor in descriptors {
            descs.push(descriptor);
        }

        let raytrace_commands = SafeCommandBuffer::new_compute(
            device,
            queue,
            1,
            &[
                // leaf node creation stage
                ComputeShaderUnit::new(
                    &[&scene_descriptor, &as_descriptor],
                    &[],
                    &RayTracingShader::acceleration_structure(
                        device,
                        raytrace_info.triangles_per_as,
                    )?,
                ),
                // tree creation stage
                ComputeShaderUnit::new(
                    &[&as_descriptor],
                    &[VkPushConstantRange::new(VK_SHADER_STAGE_COMPUTE_BIT, 0, 16)],
                    &RayTracingShader::acceleration_tree(device)?,
                ),
                // raytracing stage itself
                ComputeShaderUnit::new(
                    &descs,
                    &[VkPushConstantRange::new(VK_SHADER_STAGE_COMPUTE_BIT, 0, 8)],
                    &RayTracingShader::raytracing(
                        device,
                        raytrace_info.max_samplers,
                        raytrace_info.stack_size,
                    )?,
                ),
            ],
        )?;

        Ok(SoftwareRayTracer {
            raytrace_descriptor,
            scene_descriptor,
            as_descriptor,

            info_buffer,

            triangles_per_as: raytrace_info.triangles_per_as,
            width: image.width(),
            height: image.height(),
            valid: false,

            as_buffer,

            raytrace_commands,
        })
    }

    fn recompile(&mut self, _descriptors: &[&dyn VkHandle<VkDescriptorSetLayout>]) -> Result<()> {
        todo!()
    }

    fn check(device: &Arc<Device>) -> bool {
        let properties = device.physical_device().properties();

        if properties.limits.maxBoundDescriptorSets < 8 {
            return false;
        }

        // true

        // disable for now: http://dimov.cloud/hodasemi/Gavania/issues/26
        false
    }

    fn resize(&mut self, image: &Arc<Image>, position_image: &Arc<Image>) -> Result<()> {
        self.width = image.width();
        self.height = image.height();

        self.raytrace_descriptor.update(&[
            DescriptorWrite::storage_images(1, &[image]),
            DescriptorWrite::storage_images(2, &[position_image]),
        ])?;

        Ok(())
    }

    fn invalidate(&mut self) {
        self.valid = false;
    }

    fn process(
        &mut self,
        buffer_recorder: &mut CommandBufferRecorder<'_>,
        descriptors: &[&Arc<DescriptorSet>],
        scene_buffer: &Arc<Buffer<VertexData>>,
        index: usize,
    ) -> Result<()> {
        if !self.valid {
            self.valid = true;
            self.scene_descriptor
                .update(&[DescriptorWrite::storage_buffers(0, &[scene_buffer])])?;
        }

        let mut cb = self.raytrace_commands.begin_headless(index)?;

        // ------- create leafs -------

        // bring acceleration structure buffer into correct access mode
        cb.buffer_barrier(
            &self.as_buffer,
            0,
            VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT,
            VK_ACCESS_MEMORY_WRITE_BIT | VK_ACCESS_MEMORY_READ_BIT,
            VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT,
        );

        cb.bind_descriptor_sets_minimal(&[&self.scene_descriptor, &self.as_descriptor]);

        let mut total_node_count = 0;

        let dispatch_count = self.acceleration_leaf_count(scene_buffer.size() as u32);
        total_node_count += dispatch_count;
        cb.dispatch(dispatch_count, 1, 1);

        // make sure all leafs are written
        cb.buffer_barrier(
            &self.as_buffer,
            VK_ACCESS_MEMORY_WRITE_BIT | VK_ACCESS_MEMORY_READ_BIT,
            VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT,
            VK_ACCESS_MEMORY_WRITE_BIT | VK_ACCESS_MEMORY_READ_BIT,
            VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT,
        );

        // ------- create acceleration tree -------
        self.raytrace_commands.next_pipeline(&mut cb)?;

        // iteratively create the tree
        let mut level_count = dispatch_count;
        let mut start_index = 1;

        loop {
            let odd = level_count % 2;
            let next_level_count = (level_count as f32 / 2.0).ceil() as u32;
            let next_level_start = start_index + level_count;

            cb.push_constants(
                VK_SHADER_STAGE_COMPUTE_BIT,
                &(start_index, odd, next_level_count, next_level_start),
            );

            cb.bind_descriptor_sets_minimal(&[&self.as_descriptor]);

            total_node_count += next_level_count;
            cb.dispatch(next_level_count, 1, 1);

            // make sure current stage is written
            cb.buffer_barrier(
                &self.as_buffer,
                VK_ACCESS_MEMORY_WRITE_BIT | VK_ACCESS_MEMORY_READ_BIT,
                VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT,
                VK_ACCESS_MEMORY_WRITE_BIT | VK_ACCESS_MEMORY_READ_BIT,
                VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT,
            );

            if next_level_count == 1 {
                break;
            }

            level_count = next_level_count;
            start_index = next_level_start;
        }

        {
            let mut mapped = self.info_buffer.map_complete()?;
            mapped[0].bvh_node_count = total_node_count;
        }

        // ------- raytracing itself -------
        let mut descs = Vec::with_capacity(descriptors.len() + 1);
        descs.push(&self.raytrace_descriptor);

        for descriptor in descriptors {
            descs.push(*descriptor);
        }

        self.raytrace_commands.next_pipeline(&mut cb)?;
        cb.bind_descriptor_sets_minimal(&descs);
        cb.dispatch(self.width, self.height, 1);

        buffer_recorder.execute_commands(&[self.raytrace_commands.end(cb, index)?]);

        Ok(())
    }
}

impl SoftwareRayTracer {
    fn create_as_buffer(device: &Arc<Device>, size: VkDeviceSize) -> Result<Arc<Buffer<ASType>>> {
        Buffer::builder()
            .set_usage(VK_BUFFER_USAGE_STORAGE_BUFFER_BIT)
            .set_memory_usage(MemoryUsage::GpuOnly)
            .set_size(size)
            .build(device.clone())
    }

    fn acceleration_leaf_count(&self, vertex_count: u32) -> u32 {
        (vertex_count as f32 / (self.triangles_per_as as f32 * 3.0)).ceil() as u32
    }

    fn create_ray_trace_descriptor(
        device: &Arc<Device>,
    ) -> Result<(Arc<Buffer<SceneInfo>>, Arc<DescriptorSet>)> {
        let layout = DescriptorSetLayout::builder()
            .add_layout_binding(
                0,
                VK_DESCRIPTOR_TYPE_STORAGE_BUFFER,
                VK_SHADER_STAGE_COMPUTE_BIT,
                0,
            )
            .add_layout_binding(
                1,
                VK_DESCRIPTOR_TYPE_STORAGE_IMAGE,
                VK_SHADER_STAGE_COMPUTE_BIT,
                0,
            )
            .add_layout_binding(
                2,
                VK_DESCRIPTOR_TYPE_STORAGE_IMAGE,
                VK_SHADER_STAGE_COMPUTE_BIT,
                0,
            )
            .add_layout_binding(
                3,
                VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER,
                VK_SHADER_STAGE_COMPUTE_BIT,
                0,
            )
            .add_layout_binding(
                4,
                VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER,
                VK_SHADER_STAGE_COMPUTE_BIT,
                0,
            )
            .build(device.clone())?;

        let pool = DescriptorPool::builder()
            .set_layout(layout)
            .build(device.clone())?;

        let set = DescriptorPool::prepare_set(&pool).allocate()?;

        let buffer = Buffer::builder()
            .set_usage(VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT)
            .set_memory_usage(MemoryUsage::CpuOnly)
            .set_data(&[SceneInfo { bvh_node_count: 0 }])
            .build(device.clone())?;

        set.update(&[DescriptorWrite::uniform_buffers(3, &[&buffer])])?;

        Ok((buffer, set))
    }

    fn create_scene_descriptor(device: &Arc<Device>) -> Result<Arc<DescriptorSet>> {
        let layout = DescriptorSetLayout::builder()
            .add_layout_binding(
                0,
                VK_DESCRIPTOR_TYPE_STORAGE_BUFFER,
                VK_SHADER_STAGE_COMPUTE_BIT,
                0,
            )
            .build(device.clone())?;

        let pool = DescriptorPool::builder()
            .set_layout(layout)
            .build(device.clone())?;

        DescriptorPool::prepare_set(&pool).allocate()
    }
}
