use crate::prelude::*;

use std::{
    sync::{Arc, Mutex},
    time::Duration,
};

use anyhow::Result;
use utilities::{
    impl_reprc,
    prelude::cgmath::{Vector3, Vector4},
};

impl_reprc!(
    #[derive(Debug)]
    pub struct RGBBuffer {
        #[assume_reprc]
        rgb: Vector3<f32>,
    }
);

impl_reprc!(
    #[derive(Debug)]
    pub struct RGBABuffer {
        #[assume_reprc]
        rgba: Vector4<f32>,
    }
);

impl RGBABuffer {
    pub fn truncate(&self) -> RGBBuffer {
        RGBBuffer {
            rgb: self.rgba.truncate(),
        }
    }
}

pub struct PositionBuffer {
    position_image: Arc<Image>,
    buffer: Mutex<Option<Vec<RGBBuffer>>>,
}

impl PositionBuffer {
    pub fn new(position_image: &Arc<Image>) -> Self {
        PositionBuffer {
            position_image: position_image.clone(),
            buffer: Mutex::new(None),
        }
    }

    pub fn update(&mut self, position_image: &Arc<Image>) -> Result<()> {
        self.position_image = position_image.clone();
        *self.buffer.lock().unwrap() = None;

        Ok(())
    }

    fn get_buffer(&self) -> Result<Vec<RGBBuffer>> {
        let buffer: Arc<Buffer<RGBABuffer>> = Buffer::builder()
            .set_memory_usage(MemoryUsage::GpuToCpu)
            .set_usage(VK_BUFFER_USAGE_TRANSFER_DST_BIT)
            .set_size((self.position_image.width() * self.position_image.height()) as VkDeviceSize)
            .build(self.position_image.device().clone())?;

        // create a new command buffer
        let command_buffer = CommandBuffer::new_primary().build(
            self.position_image.device().clone(),
            self.position_image.queue().clone(),
        )?;

        {
            let mut buffer_recorder = command_buffer.begin(VkCommandBufferBeginInfo::new(
                VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT,
            ))?;

            let image_layout = self.position_image.image_layout();

            buffer_recorder.image_barrier_auto_stage(
                &self.position_image,
                image_layout,
                VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
            );

            buffer_recorder.copy_image_to_buffer(
                &self.position_image,
                VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
                &buffer,
                &[VkBufferImageCopy {
                    bufferOffset: 0,
                    bufferRowLength: 0,
                    bufferImageHeight: 0,
                    imageSubresource: self.position_image.full_resource_layers(),
                    imageOffset: VkOffset3D { x: 0, y: 0, z: 0 },
                    imageExtent: VkExtent3D {
                        width: self.position_image.width(),
                        height: self.position_image.height(),
                        depth: 1,
                    },
                }],
            );

            buffer_recorder.image_barrier_auto_stage(
                &self.position_image,
                VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
                image_layout,
            );
        }

        // submit current queue
        let submit = SubmitInfo::default().add_command_buffer(&command_buffer);
        let fence = Fence::builder().build(self.position_image.device().clone())?;

        self.position_image
            .queue()
            .lock()
            .unwrap()
            .submit(Some(&fence), &[submit])?;

        self.position_image
            .device()
            .wait_for_fences(&[&fence], true, Duration::from_secs(1))?;

        let data = buffer
            .map_complete()?
            .iter()
            .map(|p| p.truncate())
            .collect();

        Ok(data)
    }

    pub fn read_pixel(&self, x: u32, y: u32) -> Result<Vector3<f32>> {
        let mut buffer = self.buffer.lock().unwrap();

        if buffer.is_none() {
            *buffer = Some(self.get_buffer()?);
        }

        Ok(buffer.as_ref().unwrap()[(x + self.position_image.width() * y) as usize].rgb)
    }
}
