use crate::prelude::*;

pub(crate) use shaderc::{
    CompileOptions, Compiler, IncludeType, ResolvedInclude, ShaderKind, SpirvVersion,
};

use anyhow::Result;
use std::sync::Arc;

pub const SHADER_OPTION_COUNT: usize = 9;

#[macro_export]
macro_rules! include_shader {
    ($path: expr) => {{
        let shader = include_str!($path);
        let file_name = std::path::Path::new($path)
            .file_name()
            .unwrap()
            .to_str()
            .unwrap();

        (shader, file_name)
    }};
}

pub enum ShadersType {
    Graphics(
        Arc<ShaderModule<shader_type::Vertex>>,
        Arc<ShaderModule<shader_type::Fragment>>,
    ),
    Compute(Arc<ShaderModule<shader_type::Compute>>),
}

impl ShadersType {
    pub fn graphics(
        &self,
    ) -> (
        &Arc<ShaderModule<shader_type::Vertex>>,
        &Arc<ShaderModule<shader_type::Fragment>>,
    ) {
        match self {
            Self::Graphics(vertex_shader, fragment_shader) => (vertex_shader, fragment_shader),
            Self::Compute(_) => panic!("requested graphics on compute shader"),
        }
    }

    pub fn compute(&self) -> &Arc<ShaderModule<shader_type::Compute>> {
        match self {
            Self::Graphics(_, _) => panic!("requested compute on graphics shader"),
            Self::Compute(compute_shader) => compute_shader,
        }
    }
}

pub struct Shaders {
    pub position: ShadersType,
    pub position_coloruv: ShadersType,
    pub position_normal: ShadersType,
    pub position_normal_bone: ShadersType,
    pub position_normal_color: ShadersType,
    pub position_normal_coloruv: ShadersType,
    pub position_normal_coloruv_bone: ShadersType,
    pub position_normal_coloruv_normaluv: ShadersType,
    pub position_normal_coloruv_normaluv_bone: ShadersType,
}

pub struct ShaderCompiler<'a> {
    compiler: Compiler,
    files: Vec<(&'a str, &'a str)>,
}

impl<'a> ShaderCompiler<'a> {
    pub fn new(files: impl IntoIterator<Item = (&'a str, &'a str)>) -> Result<ShaderCompiler<'a>> {
        Ok(ShaderCompiler {
            compiler: Self::create_compiler()?,
            files: files.into_iter().collect(),
        })
    }

    pub fn compile_stages(
        &self,
        compute: bool,
        device: &Arc<Device>,
        default_options: &[(&str, Option<String>)],
        options: [&[&str]; SHADER_OPTION_COUNT],
    ) -> Result<Shaders> {
        if compute {
            assert_eq!(self.files.len(), 1);
        } else {
            assert_eq!(self.files.len(), 2);
        }

        let mut shaders = Vec::new();

        for i in 0..SHADER_OPTION_COUNT {
            let mut compile_options: Vec<(&str, Option<String>)> = Vec::new();

            for default_option in default_options {
                compile_options.push(default_option.clone());
            }

            for option in options[i] {
                compile_options.push((option, None));
            }

            let s = if compute {
                ShadersType::Compute(self.compile::<_, 0>(
                    device,
                    compile_options.as_slice(),
                    |_| {},
                )?)
            } else {
                ShadersType::Graphics(
                    self.compile::<_, 0>(device, compile_options.as_slice(), |_| {})?,
                    self.compile::<_, 1>(device, compile_options.as_slice(), |_| {})?,
                )
            };

            shaders.push(s);
        }

        Ok(Shaders {
            position: shaders.remove(0),
            position_coloruv: shaders.remove(0),
            position_normal: shaders.remove(0),
            position_normal_bone: shaders.remove(0),
            position_normal_color: shaders.remove(0),
            position_normal_coloruv: shaders.remove(0),
            position_normal_coloruv_bone: shaders.remove(0),
            position_normal_coloruv_normaluv: shaders.remove(0),
            position_normal_coloruv_normaluv_bone: shaders.remove(0),
        })
    }

    fn create_compiler() -> Result<Compiler> {
        match Compiler::new() {
            Some(compiler) => Ok(compiler),
            None => Err(anyhow::Error::msg(
                "Shaderc Error: failed creating compiler",
            )),
        }
    }
}

pub trait ShaderTypeConverter<T: ShaderType> {
    fn convert_shader_type() -> ShaderKind;
    fn compiler(&self) -> &Compiler;
    fn files(&self) -> &Vec<(&str, &str)>;

    fn create_compile_options<'b>() -> Result<CompileOptions<'b>> {
        match CompileOptions::new() {
            Some(compile_options) => Ok(compile_options),
            None => Err(anyhow::Error::msg(
                "Shaderc Error: failed creating compile-options",
            )),
        }
    }

    fn compile_module(
        &self,
        device: &Arc<Device>,
        compile_options: &CompileOptions<'_>,
        shader_text: &str,
        shader_file_name: &str,
    ) -> Result<Arc<ShaderModule<T>>> {
        let shader_kind = Self::convert_shader_type();

        let spirv = self.compiler().compile_into_spirv(
            shader_text,
            shader_kind,
            shader_file_name,
            "main",
            Some(compile_options),
        )?;

        let shader = ShaderModule::from_slice(device.clone(), spirv.as_binary_u8())?;

        Ok(shader)
    }

    fn compile_with_includes<F, E>(
        &self,
        device: &Arc<Device>,
        options: &[(&str, Option<&str>)],
        f: F,
        e: E,
    ) -> Result<Vec<Arc<ShaderModule<T>>>>
    where
        F: Fn(&str, IncludeType, &str, usize) -> std::result::Result<ResolvedInclude, String>,
        E: Fn(&mut CompileOptions<'_>),
    {
        let mut compile_options = Self::create_compile_options()?;

        compile_options.set_include_callback(f);
        e(&mut compile_options);

        for (option, value) in options {
            compile_options.add_macro_definition(option, *value);
        }

        let mut result = Vec::new();

        for (file, file_name) in self.files().iter() {
            result.push(self.compile_module(device, &compile_options, file, file_name)?);
        }

        Ok(result)
    }

    fn compile<F, const I: usize>(
        &self,
        device: &Arc<Device>,
        options: &[(&str, Option<String>)],
        f: F,
    ) -> Result<Arc<ShaderModule<T>>>
    where
        F: Fn(&mut CompileOptions<'_>),
    {
        let mut compile_options = Self::create_compile_options()?;

        for (option, value) in options {
            compile_options.add_macro_definition(option, value.as_deref());
        }

        f(&mut compile_options);

        let (file, file_name) = &self.files()[I];

        self.compile_module(device, &compile_options, file, file_name)
    }
}

macro_rules! impl_shader_converter {
    ($ss_type:ident, $sk:ident) => {
        impl<'a> ShaderTypeConverter<shader_type::$ss_type> for ShaderCompiler<'a> {
            fn convert_shader_type() -> ShaderKind {
                ShaderKind::$sk
            }

            fn compiler(&self) -> &Compiler {
                &self.compiler
            }

            fn files(&self) -> &Vec<(&'a str, &'a str)> {
                &self.files
            }
        }
    };
}

impl_shader_converter!(Vertex, Vertex);
impl_shader_converter!(Fragment, Fragment);
impl_shader_converter!(Geometry, Geometry);
impl_shader_converter!(TesselationEvaluation, TessEvaluation);
impl_shader_converter!(TesselationControl, TessControl);
impl_shader_converter!(Compute, Compute);
impl_shader_converter!(Miss, Miss);
impl_shader_converter!(RayGeneration, RayGeneration);
impl_shader_converter!(AnyHit, AnyHit);
impl_shader_converter!(ClosestHit, ClosestHit);
impl_shader_converter!(Intersection, Intersection);
